class AhoCorasickNode:
    def __init__(self):
        self.child_nodes = {}
        self.parent_node = None
        self.fail_node = None
        self.key = None
        self.value = None
        self.depth = -1

    def get_child_nodes(self):
        return self.child_nodes

    def set_child_nodes(self, children):
        self.child_nodes = children

    def get_parent_node(self):
        return self.parent_node

    def set_parent_node(self, node):
        self.parent_node = node

    def get_fail_node(self):
        return self.fail_node

    def set_fail_node(self, node):
        self.fail_node = node

    def get_depth(self):
        return self.depth

    def set_depth(self, depth):
        self.depth = depth

    def get_key(self):
        return self.key

    def set_key(self, key):
        self.key = key

    def get_value(self):
        return self.value

    def set_value(self, value):
        self.value = value

    def __str__(self):
        return f"child_nodes={self.child_nodes}, parent_node={self.parent_node}, fail_node={self.fail_node}, " \
               f"key={self.key}, value={self.value}, depth={self.depth} "



# from ..training.corpus_builder import CorpusBuilder
from ..training.model_builder import ModelBuilder

# corpus_builder = CorpusBuilder()
# # todo : 트레이닝 데이터 위치 ( 실제로는 바이너리 파일만 제공 될 예정 )
# corpus_builder.build_path("/Users/shinjunsoo/shineware/data/komoran_training_data", ".refine.txt")
# corpus_builder.save("corpus_build")

model_builder = ModelBuilder()
model_builder.build_path("corpus_build")
model_builder.save("../model")



GRAMMAR = "grammar.in"
WORD_DIC = "dic.word"
IRREGULAR_DIC = "dic.irregular"
POS_TABLE = "pos.table"
OBSERVATION = "observation.model"
TRANSITION = "transition.model"
IRREGULAR_MODEL = "irregular.model"



class Dictionary:
    def __init__(self, init_dic=None):
        if init_dic is None:
            init_dic = {}
        self._dictionary = init_dic

    def load(self, filename):
        self._dictionary = {}
        with open(filename, 'r') as f:
            for line in f.readlines():
                line = line.strip()
                if len(line) == 0:
                    continue
                _tokens = line.split('\t')
                _key = _tokens[0]
                _value_freq_dic = {}
                for _token_idx in range(1, len(_tokens)):
                    _token = _tokens[_token_idx]
                    _separator_idx = _token.rindex(':')
                    _value = _token[0:_separator_idx]
                    _freq = int(_token[_separator_idx + 1:])
                    _value_freq_dic[_value] = _freq
                self._dictionary[_key] = _value_freq_dic

    def save(self, filename):
        with open(filename, 'w') as f:
            for _key, _value_freq_dic in self._dictionary.items():
                f.write(
                    _key + "\t" + "\t".join([f"{_value}:{_freq}" for _value, _freq in _value_freq_dic.items()]) + "\n")

    def append(self, key, value, freq=1):
        _value_freq_dic = self._dictionary.get(key)
        if _value_freq_dic is None:
            _value_freq_dic = {}

        _freq = _value_freq_dic.get(value)
        if _freq is None:
            _freq = 0
        _freq += freq
        _value_freq_dic[value] = _freq
        self._dictionary[key] = _value_freq_dic

    def get_pos_freq_dict(self, key):
        return self._dictionary.get(key)

    def get_dict(self):
        return self._dictionary

    def print(self):
        print(self._dictionary)

    def __repr__(self):
        return f'Dictionary: {self._dictionary}'



import sys

from ..parser.problem_answer_info import ProblemAnswerInfo


class CorpusParser:

    def __init__(self):
        pass

    def parse(self, line):
        _problem_answers = line.split("\t")
        if len(_problem_answers) != 2:
            sys.stderr.write("Corpus format error. (CorpusParser.parse): " + line + "\n")
            sys.exit(-1)

        self._problem = _problem_answers[0]
        self._answer = _problem_answers[1]
        self._answer_list = []
        if self._parser_answer(self._answer):
            return ProblemAnswerInfo(self._problem, self._answer, self._answer_list)
        else:
            return None

    def _parser_answer(self, answer_texts):
        # todo : 현재는 multi word를 하나의 단어로 인식하는 기능은 제외 시킴
        _tokens = answer_texts.split(' ')
        for _token in _tokens:
            _split_idx = _token.rindex('/')
            _word = _token[0:_split_idx].strip()
            _pos = _token[_split_idx + 1:].strip()
            if len(_word) == 0 or len(_pos) == 0 or _pos == 'NA':
                sys.stderr.write("Corpus format error (CorpusParser._parser_answer): " + answer_texts + "\n")
                return False
            self._answer_list.append((_word, _pos))
        return True

#
# with open('/Users/shinjunsoo/shineware/data/komoran_training_data/현대문어_형태분석_말뭉치.refine.txt') as f:
#     corpus_parser = CorpusParser()
#     for line in f.readlines():
#         line = line.strip()
#         if len(line) == 0:
#             continue



class StringUtil:

    @staticmethod
    def is_basic_latic(hex_code):
        return 0x0000 <= hex_code <= 0x007F

    @staticmethod
    def is_num(hex_code):
        return 0x0030 <= hex_code <= 0x0039

    @staticmethod
    def is_english(hex_code):
        return 0x0041 <= hex_code <= 0x005A or 0x0061 <= hex_code <= 0x007A

    @staticmethod
    def is_korean(hex_code):
        # HANGUL SYLLABLES
        # or HANGUL COMPATIBILITY JAMO
        # or HANGUL JAMO
        return 0xAC00 <= hex_code <= 0xD7A3 \
               or 0x3131 <= hex_code <= 0x318E \
               or 0x1100 <= hex_code <= 0x11F9

    @staticmethod
    def is_japanese(hex_code):
        # KATAKANA
        # or KATAKANA PHONETIC EXTENSIONS
        # or HIRAGANA
        return 0x30A0 <= hex_code <= 0x30FF \
               or 0x31F0 <= hex_code <= 0x31FF \
               or 0x3041 <= hex_code <= 0x309F

    @staticmethod
    def is_chinese(hex_code):
        # CJK_COMPATIBILITY
        # or CJK_UNIFIED_IDEOGRAPHS
        # or CJK_UNIFIED_IDEOGRAPHS_EXTENSION_A
        # or CJK_UNIFIED_IDEOGRAPHS_EXTENSION_B
        # or CJK_COMPATIBILITY_IDEOGRAPHS
        return 0x3300 <= hex_code <= 0x33FE \
               or 0x4E00 <= hex_code <= 0x9FFF \
               or 0x3400 <= hex_code <= 0x4DB5 \
               or 0x20000 <= hex_code <= 0x2A6DF \
               or 0xF900 <= hex_code <= 0xFACF



import math
import os
import shutil

from ..common.dictionary import Dictionary
from ..common.grammar import Grammar
from ..common.irregular.irregular_node import IrregularNode

from ..common.irregular.irregular_trie import IrregularTrie
from ..constant import SYMBOL, FILENAME
from ..parser.korean_unit_parser import KoreanUnitParser
from ..training.model.observation import Observation
from ..training.model.pos_table import PosTable
from ..training.model.transition import Transition


class ModelBuilder:
    def __init__(self):
        self.unit_parser = KoreanUnitParser()
        self.word_dic = Dictionary()
        self.irr_dic = Dictionary()
        self.grammar = Grammar()
        self.pos_table = PosTable()
        self.transition = Transition(0)
        self.observation = Observation()

    def build_path(self, path):
        self.__init__()
        self.word_dic.load(path + "/" + FILENAME.WORD_DIC)
        # todo : add external dic 부분은 구현하지 않음
        self.irr_dic.load(path + "/" + FILENAME.IRREGULAR_DIC)
        self.grammar.load(path + "/" + FILENAME.GRAMMAR)

        total_prev_pos_count = self.__get_total_prev_pos_count__()
        self._build_pos_table_(total_prev_pos_count)
        self._cal_transition_(total_prev_pos_count)
        self._cal_observation_(total_prev_pos_count)
        self._build_irregular_dic()

    def __get_total_prev_pos_count__(self):
        _pos_count_dic = {}
        for prev_pos in self.grammar.get_grammar().keys():
            cur_pos_dic = self.grammar.get_grammar()[prev_pos]
            for _cur_pos in cur_pos_dic.keys():
                _freq = 0
                if prev_pos in _pos_count_dic:
                    _freq = _pos_count_dic[prev_pos]
                _freq += cur_pos_dic[_cur_pos]
                _pos_count_dic[prev_pos] = _freq

        return _pos_count_dic

    def _build_pos_table_(self, _total_prev_pos_count):
        self.pos_table.__init__()
        for pos in _total_prev_pos_count.keys():
            self.pos_table.put(pos)
        self.pos_table.put(SYMBOL.BOE)
        self.pos_table.put(SYMBOL.EOE)
        self.pos_table.put(SYMBOL.NA)

    def _cal_transition_(self, total_prev_pos_count):
        self.transition.__init__(self.pos_table.size())
        for prev_pos in total_prev_pos_count.keys():
            cur_pos_dic = self.grammar.get_grammar().get(prev_pos)
            for cur_pos in cur_pos_dic:
                prev_to_cur_freq = cur_pos_dic.get(cur_pos)
                prev_freq = total_prev_pos_count.get(prev_pos)
                # todo : NNP인 경우에 빈도수를 뻥튀기 하는 부분은 구현하지 않음
                transition_score = prev_to_cur_freq / prev_freq
                transition_score = math.log(transition_score)
                prev_pos_id = self.pos_table.get_id(prev_pos)
                cur_pos_id = self.pos_table.get_id(cur_pos)
                self.transition.put(prev_pos_id, cur_pos_id, transition_score)
        self.transition.put(self.pos_table.get_id(SYMBOL.NA), self.pos_table.get_id(SYMBOL.EOE), -10000.0)

    def _cal_observation_(self, total_prev_pos_count):
        self.observation = Observation()
        for word in self.word_dic.get_dict():
            pos_freq_dic = self.word_dic.get_pos_freq_dict(word)
            for pos in pos_freq_dic:
                total_count = total_prev_pos_count.get(pos)
                if total_count is None:
                    print(f"{pos}가 prev pos count에 없습니다!")
                observation_score = pos_freq_dic[pos] / total_count
                observation_score = math.log(observation_score)
                self.observation.put(word, pos, self.pos_table.get_id(pos), observation_score)

    def save(self, path):
        if os.path.exists(path):
            shutil.rmtree(path)
        os.makedirs(path)
        self.transition.save(path + "/" + FILENAME.TRANSITION)
        self.observation.save(path + "/" + FILENAME.OBSERVATION)
        self.pos_table.save(path + "/" + FILENAME.POS_TABLE)
        self.irr_trie.save(path + "/" + FILENAME.IRREGULAR_MODEL)

    def _build_irregular_dic(self):
        self.irr_trie = IrregularTrie()
        for irr_pattern in self.irr_dic.get_dict():
            irr_pattern_jaso = self.unit_parser.parse(irr_pattern)
            convert_rule_freq_dic = self.irr_dic.get_dict().get(irr_pattern)
            for convert_rule in convert_rule_freq_dic:
                # todo : KOMORAN에서는 빈도수로 pruning 하는 로직이 있음
                convert_freq = convert_rule_freq_dic.get(convert_rule)
                irr_node = self.make_irr_node(irr_pattern_jaso, self.unit_parser.parse(convert_rule))
                self.irr_trie.put(irr_pattern_jaso, irr_node)

    def make_irr_node(self, irr_pattern_jaso, convert_rule_jaso):
        # todo : KOMORAN 코드 구조에서 보면 first pos id와 irregular tokens 정보만 있으면 되는 것으로 보임
        irr_node = IrregularNode()
        irr_tokens = []
        tokens = convert_rule_jaso.split(' ')
        for idx, token in enumerate(tokens):
            morph, pos = token.rsplit('/', 1)
            pos_id = self.pos_table.get_id(pos)
            if idx == 0:
                irr_node.set_first_pos_id(pos_id)
            irr_tokens.append((morph, pos_id))
        irr_node.set_tokens(irr_tokens)
        return irr_node



import time
from multiprocessing.pool import Pool

from ..core.komoran import Komoran

komoran = Komoran("../training/komoran_model")
komoran.set_fwd("../fwd.dic")
komoran.set_user_dic("../user.dic")

f = open("../core/stress.test", "r")
cnt = 0


def analyze(_line):
    return komoran.analyze(_line.strip())


start = time.time()  # 시작 시간 저장

pool = Pool(processes=8)
analyze_results = pool.map(analyze, f.readlines())
for analyze_result in analyze_results:
    result_tmp = []
    for word, pos in analyze_result:
        result_tmp.append(word+"/"+pos)
    print(f"{cnt}={' '.join(result_tmp)}")
    cnt += 1
pool.close()
pool.join()
end = time.time()  # 시작 시간 저장
print(f"elapsed time={(end - start)}")

f.close()



class IrregularNode:

    def get_first_pos_id(self):
        return self.first_pos_id

    def set_first_pos_id(self, pos_id):
        self.first_pos_id = pos_id

    def set_tokens(self, irr_node_tokens):
        self.irr_node_tokens = irr_node_tokens

    def get_tokens(self):
        return self.irr_node_tokens

    def __eq__(self, other):
        if isinstance(other, IrregularNode):
            return self.first_pos_id == other.first_pos_id and self.irr_node_tokens == other.irr_node_tokens
        return False

    def __repr__(self):
        return f"(first_pos_id={self.first_pos_id}, irr_node_tokens={self.irr_node_tokens})"




from ...common.ahocorasick.ahocorasick_dictionary import AhoCorasickDictionary
from ...parser.korean_unit_parser import KoreanUnitParser


class Observation:
    def __init__(self):
        self.ahocorasick = AhoCorasickDictionary()
        self.parser = KoreanUnitParser()

    def put(self, word, pos, pos_id, observation_score):
        korean_units = self.parser.parse(word)
        score_pos_list = self.ahocorasick.get_value(korean_units)
        if score_pos_list is None:
            # todo : KOMORAN에서는 score tag 클래스로 아래 3가지 정보를 담고 있음
            score_pos_list = [(pos, pos_id, observation_score)]
        else:
            has_same_pos = False
            for _pos, _pos_id, _observation_score in score_pos_list:
                if _pos_id == pos_id:
                    has_same_pos = True
                    break
            if not has_same_pos:
                score_pos_list.append((pos, pos_id, observation_score))
        self.ahocorasick.put(korean_units, score_pos_list)

    def get_dictionary(self):
        return self.ahocorasick

    def save(self, filename):
        self.ahocorasick.save(filename)

    def load(self, filename):
        self.ahocorasick.load(filename)



class PosTable:
    def __init__(self):
        self.pos_id_table = {}
        self.id_pos_table = {}

    def put(self, pos):
        if pos not in self.pos_id_table:
            self.pos_id_table[pos] = len(self.pos_id_table)
            self.id_pos_table[len(self.id_pos_table)] = pos

    def get_id(self, pos):
        return self.pos_id_table.get(pos)

    def get_pos(self, pos_id):
        return self.id_pos_table.get(pos_id)

    def size(self):
        return len(self.pos_id_table)

    def save(self, filename):
        with open(filename, 'w') as f:
            for pos, pos_id in self.pos_id_table.items():
                f.write(f"{pos}\t{pos_id}\n")
        # todo : 세종 테그셋 id를 빌드하는 부분은 생략 됨

    def load(self, filename):
        with open(filename, 'r') as f:
            self.__init__()
            for line in f.readlines():
                line = line.strip()
                if len(line) == 0:
                    continue
                pos, pos_id = line.split('\t')
                self.pos_id_table[pos] = int(pos_id)
                self.id_pos_table[int(pos_id)] = pos
        # todo : 세종 테그셋 id를 빌드하는 부분은 생략 됨


# pos_table = PosTable()
# pos_table.put('NNG')
# pos_table.put('NNP')
# pos_table.put('NNG')
# pos_table.put('NNK')
# pos_table.save("wow.table")
# pos_table.load("wow.table")



import math

from ..common.irregular.irregular_trie import IrregularTrie
from ..constant import SYMBOL, FILENAME
from ..parser.korean_unit_parser import KoreanUnitParser
from ..training.model.observation import Observation
from ..training.model.pos_table import PosTable
from ..training.model.transition import Transition
from ..utils.string_util import StringUtil


class Model:
    def __init__(self, model_path):
        self.transition = Transition(0)
        self.observation = Observation()
        self.pos_table = PosTable()
        self.irr_trie = IrregularTrie()
        self.transition.load(model_path + "/" + FILENAME.TRANSITION)
        self.observation.load(model_path + "/" + FILENAME.OBSERVATION)
        self.pos_table.load(model_path + "/" + FILENAME.POS_TABLE)
        self.irr_trie.load(model_path + "/" + FILENAME.IRREGULAR_MODEL)
        self.observation.get_dictionary().build_fail_link()
        self.irr_trie.get_dictionary().build_fail_link()


class LatticeNode:
    def __init__(self, begin_idx, end_idx, word, pos, pos_id, score):
        self.__begin_idx = begin_idx
        self.__end_idx = end_idx
        self.__word = word
        self.__pos = pos
        self.__pos_id = pos_id
        self.__score = score
        self.__prev_node_idx = -1
        self.__is_irregular = False

    def __repr__(self):
        return f"""(begin_idx={self.__begin_idx}, end_idx={self.end_idx}, word={self.word}, pos={self.pos}, pos_id={self.pos_id}, score={self.score}, prev_node_idx={self.prev_node_idx}, is_irregular={self.__is_irregular})"""

    @property
    def begin_idx(self):
        return self.__begin_idx

    @begin_idx.setter
    def begin_idx(self, value):
        self.__begin_idx = value

    @property
    def end_idx(self):
        return self.__end_idx

    @end_idx.setter
    def end_idx(self, value):
        self.__end_idx = value

    @property
    def word(self):
        return self.__word

    @word.setter
    def word(self, value):
        self.__word = value

    @property
    def pos(self):
        return self.__pos

    @pos.setter
    def pos(self, value):
        self.pos = value

    @property
    def pos_id(self):
        return self.__pos_id

    @pos_id.setter
    def pos_id(self, value):
        self.__pos_id = value

    @property
    def score(self):
        return self.__score

    @score.setter
    def score(self, value):
        self.__score = value

    @property
    def prev_node_idx(self):
        return self.__prev_node_idx

    @prev_node_idx.setter
    def prev_node_idx(self, value):
        self.__prev_node_idx = value

    @property
    def is_irregular(self):
        return self.__is_irregular

    @is_irregular.setter
    def is_irregular(self, value):
        self.__is_irregular = value


class Lattice:
    def __init__(self, model, user_dic):
        self.model = model
        self.user_dic = user_dic
        self.init_node()
        self.make_new_contexts()

    """
    word_pos_list = [('ㄱㅗ', 15), (',', 5)]
    """

    def attach_info_to_word_pos_list(self, word_pos_list):
        results = []
        for word, pos_id in word_pos_list:
            # return get_value 의 형태는 이런 식임 [('EC', 15, -2.107718373555622), ('EF', 19, -3.972572410841921)]
            # todo : for 문을 돌면서 list에서 pos에 해당하는 score 값을 찾아야 하기 때문에 이 부분은 추후 개선의 여지가 필요함 ( 바로 찾을 수 있는 방법.. )
            # observation score 가져오기
            for _pos, _pos_id, _score in self.model.observation.get_dictionary().get_value(word):
                if pos_id == _pos_id:
                    results.append((word, _pos, pos_id, _score))
                    break
        return results

    def make_new_contexts(self):
        self.observation_context = self.model.observation.get_dictionary().new_find_context()
        self.irregular_context = self.model.irr_trie.get_dictionary().new_find_context()
        if self.user_dic is not None:
            self.user_dic_context = self.user_dic.get_dictionary().new_find_context()

    def init_node(self):
        self.lattice = {}
        self.irr_idx = 0
        lattice_nodes = [LatticeNode(-1, 0, SYMBOL.BOE, SYMBOL.BOE, self.model.pos_table.get_id(SYMBOL.BOE), 0.0)]
        self.lattice[0] = lattice_nodes

    def get_observation(self, jaso_unit):
        return self.model.observation.get_dictionary().get(jaso_unit, self.observation_context)

    def get_irregular_nodes(self, jaso_unit):
        return self.model.irr_trie.get_dictionary().get(jaso_unit, self.irregular_context)

    def get_user_dic(self, jaso_unit):
        if self.user_dic is not None:
            return self.user_dic.get_dictionary().get(jaso_unit, self.user_dic_context)
        return None

    def put(self, begin_idx, end_idx, word, pos, pos_id, observation_score, is_irregular=False):
        prev_lattice_nodes = self.lattice.get(begin_idx)
        if prev_lattice_nodes is None:
            return False
        else:
            max_lattice_node = self.get_max_transition_node(prev_lattice_nodes, begin_idx, end_idx, word, pos, pos_id,
                                                            observation_score)
            if max_lattice_node is not None:
                max_lattice_node.is_irregular = is_irregular
                self.append_node(max_lattice_node)
                return True
        return False

    def get_max_transition_node(self, prev_lattice_nodes, begin_idx, end_idx, word, pos, pos_id, observation_score):
        prev_max_node = None
        max_score = -math.inf
        lattice_node_idx = -1
        prev_lattice_node_idx = -1
        for prev_lattice_node in prev_lattice_nodes:
            lattice_node_idx += 1
            # 불규칙 확장인 경우
            # todo : KOMORAN에서는 pos_id == -1 인 경우(불규칙 확장. 일반 불규칙이 아님)에 continue 했으나 해당 로직이 왜 필요한지는 의문...
            # 일단 KOMORAN 의 불규칙 확장 알고리즘 중 자식 노드 유무에 따라 탐색 후보로 등록하는 부분이 제외되었기 때문에 아래 continue 부분은 실제로 걸리는 케이스가 없음
            # if prev_lattice_node.is_irregular:
            if prev_lattice_node.pos_id == -1:
                continue

            prev_pos_id = -1
            prev_word = ''
            if prev_lattice_node.pos == SYMBOL.EOE:
                prev_pos_id = self.model.pos_table.get_id(SYMBOL.BOE)
                prev_word = SYMBOL.BOE
            else:
                prev_pos_id = prev_lattice_node.pos_id
                prev_word = prev_lattice_node.word

            # 전이확률 값
            transition_score = self.model.transition.get(prev_pos_id, pos_id)
            if transition_score is None:
                continue

            # todo : 결합 규칙 체크 부분은 생략 됨

            prev_observation_score = prev_lattice_node.score
            if max_score < transition_score + prev_observation_score:
                max_score = transition_score + prev_observation_score
                prev_max_node = prev_lattice_node
                prev_lattice_node_idx = lattice_node_idx
        if prev_max_node is not None:
            lattice_node = LatticeNode(begin_idx, end_idx, word, pos, pos_id, max_score + observation_score)
            lattice_node.prev_node_idx = prev_lattice_node_idx
            return lattice_node

        return None

    def append_node(self, lattice_node):
        lattice_nodes = self.lattice.get(lattice_node.end_idx)
        if lattice_nodes is None:
            lattice_nodes = []
        lattice_nodes.append(lattice_node)
        self.lattice[lattice_node.end_idx] = lattice_nodes
        return len(lattice_nodes) - 1


class ChunkLetter:
    def __init__(self):
        self.prev_pos = ""
        self.prev_word = ""
        self.prev_begin_idx = 0

    def get_prev_pos(self):
        return self.prev_pos

    def get_prev_word(self):
        return self.prev_word

    def get_prev_begin_idx(self):
        return self.prev_begin_idx

    def set_prev_pos(self, prev_pos):
        self.prev_pos = prev_pos

    def set_prev_word(self, prev_word):
        self.prev_word = prev_word

    def set_prev_begin_idx(self, prev_begin_idx):
        self.prev_begin_idx = prev_begin_idx


class Komoran:
    def __init__(self, model_path):
        print(f'load model : {model_path}')
        self.model = Model(model_path)
        self.unit_parser = KoreanUnitParser()
        self.user_dic = None
        self.fwd = None
        print(f'load done')

    def set_user_dic(self, user_dic_path):
        self.user_dic = Observation()
        with open(user_dic_path, 'r') as f:
            for line in f.readlines():
                line = line.strip()
                if len(line) == 0 or line[0] == '#':
                    continue
                tokens = line.split('\t', 1)
                if len(tokens) == 1:
                    word = tokens[0]
                    pos = "NNP"
                else:
                    word = tokens[0]
                    pos = tokens[1]
                self.user_dic.put(word, pos, self.model.pos_table.get_id(pos), 0.0)
        self.user_dic.get_dictionary().build_fail_link()

    def set_fwd(self, fwd_path):
        with open(fwd_path, 'r') as f:
            self.fwd = {}
            for line in f.readlines():
                line = line.strip()
                if len(line) == 0 or line[0] == '#':
                    continue
                tokens = line.split('\t')
                if len(tokens) != 2:
                    continue
                target_word = tokens[0]
                analyzed_results = []
                for analyzed_result in tokens[1].split(' '):
                    word, pos = analyzed_result.rsplit('/', 1)
                    analyzed_results.append((word, pos))
                self.fwd[self.unit_parser.parse(target_word)] = analyzed_results

    def analyze(self, sentence):
        lattice = Lattice(self.model, self.user_dic)
        jaso_units = self.unit_parser.parse(sentence)
        extra_lattice_idx = -1
        whitespace_idx = 0
        idx = 0
        chunk_letter = ChunkLetter()
        while idx < len(jaso_units):
            jaso_unit = jaso_units[idx]
            # todo : here we go! 기호 처리 결과 확인 및 연속 기호 처리
            # 기분석 사전 적용
            next_whitespace_idx, extra_lattice_idx = self.lookup_fwd(lattice, jaso_units, idx, extra_lattice_idx)
            if next_whitespace_idx != -1:
                idx = next_whitespace_idx - 1
                continue
            # 공백 문자인 경우에
            if jaso_unit == ' ':
                self.consume_remain_chunk_letter(lattice, idx, chunk_letter)
                self.bridging_lattice(lattice, whitespace_idx, idx, jaso_units)
                whitespace_idx = idx + 1

            # 사용자 사전 적용
            self.user_dic_parsing(lattice, jaso_unit, idx)
            # 기호 파싱
            self.symbol_parsing(lattice, jaso_unit, idx)
            # 연속 문자 (숫자, 영어, 한자 등)
            self.chunk_letter_parsing(lattice, jaso_unit, idx, chunk_letter)

            # 단어 파싱
            self.regular_parsing(lattice, jaso_unit, idx)
            extra_lattice_idx = self.irregular_parsing(lattice, jaso_unit, idx, extra_lattice_idx)
            self.irregular_extends(lattice, jaso_unit, idx)
            idx += 1

        last_idx = len(jaso_units)
        # 단어 끝에 어절 마지막 노드 추가
        self.consume_remain_chunk_letter(lattice, last_idx, chunk_letter)
        self.bridging_lattice(lattice, whitespace_idx, last_idx, jaso_units)
        last_idx += 1

        # for i in range(irr_idx, last_idx + 1):
        #     lattice_nodes = lattice.lattice.get(i)
        #     if lattice_nodes is None:
        #         continue
        #     print(f'{i} :')
        #     for lattice_node in lattice_nodes:
        #         print(f'\t{lattice_node}')

        # get max transitions
        result = []
        end_node = lattice.lattice.get(last_idx)[0]
        begin_idx = end_node.begin_idx
        prev_node_idx = end_node.prev_node_idx
        while True:
            node = lattice.lattice.get(begin_idx)[prev_node_idx]
            if node.word == SYMBOL.BOE:
                break
            word = node.word
            pos = node.pos
            result.append((word, pos))
            begin_idx = node.begin_idx
            prev_node_idx = node.prev_node_idx
        return reversed(result)
        # for word, pos in reversed(result):
        #     print(f"{word}/{pos}")

    def regular_parsing(self, lattice, jaso_unit, idx):
        # 아래와 같은 형태가 리턴 됨
        # {
        #   'ㅈㅏㅈ': [('VA', 20, -3.2449828635675546), ('VV', 14, -5.561212329330546)],
        #   'ㅈ': [('NNG', 2, -4.833270674064261)]
        # }
        word_with_pos_scores = lattice.get_observation(jaso_unit)
        if len(word_with_pos_scores) == 0:
            return
        for word, pos_scores in word_with_pos_scores.items():
            begin_idx = idx - len(word) + 1
            end_idx = idx + 1
            for pos, pos_id, observation_score in pos_scores:
                lattice.put(begin_idx, end_idx, word, pos, pos_id, observation_score)

    def irregular_parsing(self, lattice, jaso_unit, idx, extra_lattice_idx):
        # get_irregular_nodes 는 아래와 같은 형태가 리턴 됨
        # {
        #   'ㄱㅏ': [('ㅇㅓ', 15), ('ㄱㅏ', 27), ('ㅇㅏ', 15)]
        # }
        word_with_irr_nodes = lattice.get_irregular_nodes(jaso_unit)
        if len(word_with_irr_nodes) == 0:
            return extra_lattice_idx
        for word, irr_nodes in word_with_irr_nodes.items():
            # print(f"{word} -> {irr_nodes}")
            begin_idx = idx - len(word) + 1
            end_idx = idx + 1
            for irr_node in irr_nodes:
                word_with_pos_scores = lattice.attach_info_to_word_pos_list(irr_node.get_tokens())
                irregular_node_tokens_size = len(word_with_pos_scores)
                if irregular_node_tokens_size == 1:
                    _word, pos, pos_id, observation_score = word_with_pos_scores[0]
                    lattice.put(begin_idx, end_idx, _word, pos, pos_id, observation_score, True)
                    extra_lattice_idx -= 1
                else:
                    cnt = 0
                    for _word, pos, pos_id, observation_score in word_with_pos_scores:
                        if cnt == 0:
                            lattice.put(begin_idx, extra_lattice_idx, _word, pos, pos_id, observation_score, True)
                        elif cnt == len(word_with_pos_scores) - 1:
                            lattice.put(extra_lattice_idx + 1, end_idx, _word, pos, pos_id, observation_score, True)
                        else:
                            lattice.put(extra_lattice_idx + 1, extra_lattice_idx, _word, pos, pos_id, observation_score,
                                        True)
                        extra_lattice_idx -= 1
                        cnt += 1
        return extra_lattice_idx

    def irregular_extends(self, lattice, jaso_unit, idx):
        # todo : KOMORAN에서는 현재 자소를 뒤에 붙여서 확장하는 방법 외에 자식 노드가 있으면 계속 탐색 가능하게 처리하는 로직이 있음.
        # 그러나 해당 로직으로 인해서 어떠한 장점이 있는지 확인되지 않음. 그렇기 때문에 KOMORANPy 에서는 생략.
        prev_lattice_nodes = lattice.lattice.get(idx)
        if prev_lattice_nodes is None:
            return
        for prev_lattice_node in prev_lattice_nodes:
            if prev_lattice_node.is_irregular:
                extended_word = prev_lattice_node.word + jaso_unit
                # get_value 리턴 형태 = [('EP', 29, -0.723657199450198), ('VV', 14, -13.721442706150867)]
                observation_values = self.model.observation.get_dictionary().get_value(extended_word)
                if observation_values is None:
                    continue
                for pos, pos_id, score in observation_values:
                    lattice.put(prev_lattice_node.begin_idx, idx + 1, extended_word, pos, pos_id, score)

    def lookup_fwd(self, lattice, jaso_unit, idx, extra_lattice_idx):
        if self.fwd is None:
            return -1, extra_lattice_idx
        if idx == 0 or jaso_unit[idx - 1] == ' ':
            word_end_idx = jaso_unit.find(' ', idx)
            word_end_idx = len(jaso_unit) if word_end_idx == -1 else word_end_idx
            target_word = jaso_unit[idx:word_end_idx]
            word_pos_pairs = self.fwd.get(target_word)
            if word_pos_pairs is None:
                return -1, extra_lattice_idx
            # put(self, begin_idx, end_idx, word, pos, pos_id, observation_score, is_irregular=False):
            length = len(word_pos_pairs)
            if length == 1:
                word, pos = word_pos_pairs[0]
                pos_id = self.model.pos_table.get_id(pos)
                lattice.put(begin_idx=idx, end_idx=word_end_idx, word=word, pos=pos, pos_id=pos_id,
                            observation_score=0.0)
            else:
                fwd_idx = 0
                for word, pos in word_pos_pairs:
                    pos_id = self.model.pos_table.get_id(pos)
                    if fwd_idx == 0:
                        lattice.put(begin_idx=idx, end_idx=extra_lattice_idx - 1, word=word, pos=pos, pos_id=pos_id,
                                    observation_score=0.0)
                    elif fwd_idx == length - 1:
                        lattice.put(begin_idx=extra_lattice_idx, end_idx=word_end_idx, word=word, pos=pos,
                                    pos_id=pos_id, observation_score=0.0)
                    else:
                        lattice.put(begin_idx=extra_lattice_idx, end_idx=extra_lattice_idx - 1, word=word, pos=pos,
                                    pos_id=pos_id, observation_score=0.0)
                    extra_lattice_idx -= 1
                    fwd_idx += 1
            return word_end_idx, extra_lattice_idx
        return -1, extra_lattice_idx

    def bridging_lattice(self, lattice, whitespace_idx, idx, jaso_units):
        end_node_inserted = lattice.put(idx, idx + 1, SYMBOL.EOE, SYMBOL.EOE, self.model.pos_table.get_id(SYMBOL.EOE),
                                        0.0)
        # 마지막 노드가 추가되지 않았다면 NA임
        if not end_node_inserted:
            na_word = jaso_units[whitespace_idx:idx]
            na_lattice_node = LatticeNode(whitespace_idx, idx, na_word, SYMBOL.NA,
                                          self.model.pos_table.get_id(SYMBOL.NA), -10000.0)
            na_lattice_node.prev_node_idx = 0
            na_node_idx = lattice.append_node(na_lattice_node)

            end_lattice_node = LatticeNode(idx, idx + 1, SYMBOL.EOE, SYMBOL.EOE,
                                           self.model.pos_table.get_id(SYMBOL.EOE), 0.0)
            end_lattice_node.prev_node_idx = na_node_idx
            lattice.append_node(end_lattice_node)

    def user_dic_parsing(self, lattice, jaso_unit, idx):
        # 아래와 같은 형태가 리턴 됨
        # {
        #   'ㅈㅏㅈ': [('VA', 20, -3.2449828635675546), ('VV', 14, -5.561212329330546)],
        #   'ㅈ': [('NNG', 2, -4.833270674064261)]
        # }
        word_with_pos_scores = lattice.get_user_dic(jaso_unit)
        if word_with_pos_scores is None or len(word_with_pos_scores) == 0:
            return
        for word, pos_scores in word_with_pos_scores.items():
            begin_idx = idx - len(word) + 1
            end_idx = idx + 1
            for pos, pos_id, observation_score in pos_scores:
                lattice.put(begin_idx, end_idx, word, pos, pos_id, observation_score)

    def symbol_parsing(self, lattice, jaso_unit, idx):
        hex_value = ord(jaso_unit)
        if StringUtil.is_num(hex_value):
            return
        elif StringUtil.is_basic_latic(hex_value):
            if StringUtil.is_english(hex_value) is False \
                    and hex_value != 0x20 \
                    and self.model.observation.ahocorasick.get_value(jaso_unit) is None:
                lattice.put(idx, idx + 1, jaso_unit, SYMBOL.SW, self.model.pos_table.get_id(SYMBOL.SW), -10000.0)
        elif StringUtil.is_korean(hex_value) is False and StringUtil.is_japanese(
                hex_value) is False and StringUtil.is_chinese(hex_value):
            lattice.put(idx, idx + 1, jaso_unit, SYMBOL.SW, self.model.pos_table.get_id(SYMBOL.SW), -10000.0)

    def chunk_letter_parsing(self, lattice, jaso_unit, idx, chunk_letter):
        cur_pos = ""
        hex_value = ord(jaso_unit)
        if StringUtil.is_english(hex_value):
            cur_pos = "SL"
        elif StringUtil.is_num(hex_value):
            cur_pos = "SN"
        elif StringUtil.is_chinese(hex_value):
            cur_pos = "SH"
        elif StringUtil.is_japanese(hex_value):
            cur_pos = "SL"

        if cur_pos == chunk_letter.get_prev_pos():
            chunk_letter.set_prev_word(chunk_letter.get_prev_word() + jaso_unit)
        else:
            if chunk_letter.get_prev_pos() == "SL" or chunk_letter.get_prev_pos() == "SH" or chunk_letter.get_prev_pos() == "SN":
                lattice.put(begin_idx=chunk_letter.get_prev_begin_idx(),
                            end_idx=idx,
                            word=chunk_letter.get_prev_word(),
                            pos=chunk_letter.get_prev_pos(),
                            pos_id=self.model.pos_table.get_id(chunk_letter.get_prev_pos()),
                            observation_score=-1.0)
            chunk_letter.set_prev_begin_idx(idx)
            chunk_letter.set_prev_word(jaso_unit)
            chunk_letter.set_prev_pos(cur_pos)

    def consume_remain_chunk_letter(self, lattice, idx, chunk_letter):
        if len(chunk_letter.get_prev_pos().strip()) != 0:
            lattice.put(begin_idx=chunk_letter.get_prev_begin_idx(),
                        end_idx=idx,
                        word=chunk_letter.get_prev_word(),
                        pos=chunk_letter.get_prev_pos(),
                        pos_id=self.model.pos_table.get_id(chunk_letter.get_prev_pos()),
                        observation_score=-1.0)

# komoran = Komoran("../training/komoran_model")
# komoran.set_fwd("../fwd.dic")
# komoran.set_user_dic("../user.dic")
# print(komoran.analyze("이번 감기는 강하다"))
# print()
# print(komoran.analyze("골렸어"))
# print()
# print(komoran.analyze("샀으니"))
# print()
# print(komoran.analyze("이어져서"))  # --> 이어지/VV 어서/EC 가 정답임
# print()
# print(komoran.analyze("러너라는"))  # --> 러너/NNG 이라는
# print()
# print(komoran.analyze("이정도로 골렸어 뷁"))
# print()
# print(komoran.analyze("뷁뷁 뷁부어"))
# print()
#
# ch = '감'
# hex_value = ord(ch)
# if 0x0000 < hex_value < 0x007E:
#     print('basic latic')
# else:
#     print('other')



import time

from ..core.komoran import Komoran

komoran = Komoran("../training/komoran_model")
komoran.set_fwd("../fwd.dic")
komoran.set_user_dic("../user.dic")

f = open("../core/stress.test", "r")

start = time.time()  # 시작 시간 저장
cnt = 0
for line in f.readlines():
    komoran.analyze(line.strip())
    if cnt % 10000 == 0:
        print(f"{cnt}")
    cnt += 1
end = time.time()  # 시작 시간 저장
print(f"elapsed time={(end - start)}")

f.close()



from ..ahocorasick.ahocorasick_dictionary import AhoCorasickDictionary


class IrregularTrie:
    def __init__(self):
        self.ahocorasick = AhoCorasickDictionary()

    def put(self, irr_pattern, irr_node):
        irr_nodes = self.ahocorasick.get_value(irr_pattern)
        if irr_nodes is None:
            irr_nodes = [irr_node]
        else:
            include_same_node = False
            for _irr_node in irr_nodes:
                if irr_node == _irr_node:
                    include_same_node = True
                    break
            if not include_same_node:
                irr_nodes.append(irr_node)
        self.ahocorasick.put(irr_pattern, irr_nodes)

    def get_dictionary(self):
        return self.ahocorasick

    def save(self, filename):
        self.ahocorasick.save(filename)

    def load(self, filename):
        self.ahocorasick.load(filename)



class ProblemAnswerInfo:
    def __init__(self, problem, answer, answer_list):
        self._problem = problem
        self._answer = answer
        self._answer_list = answer_list

    def get_problem(self):
        return self._problem

    def get_answer(self):
        return self._answer

    def get_answer_list(self):
        return self._answer_list

    def __str__(self):
        return f"[ProblemAnswerInfo] problem={self._problem}, answer={self._answer}, answer_list={self._answer_list}"



from ..parser.korean_unit_parser import is_korean


class IrregularParser:
    def __init__(self):
        self._answer_begin_idx = -1
        self._problem_begin_idx = -1
        self._answer_end_idx = -1
        self._problem_end_idx = -1

    def parse(self, problem, answer_list):
        self.__init__()
        self._set_begin_idx(problem, answer_list)
        self._set_end_idx(problem, answer_list)
        return self._get_irregular_rules(problem, answer_list)

    def _set_begin_idx(self, problem, answer_list):
        # left to right retrieval
        _ptr = 0
        for _idx, _answer in enumerate(answer_list):
            _word, _pos = _answer
            if _ptr + len(_word) > len(problem):
                self._answer_begin_idx = _idx
                self._problem_begin_idx = _ptr
                break

            _problem_sub_token = problem[_ptr:_ptr + len(_word)]
            if _word != _problem_sub_token:
                self._answer_begin_idx = _idx
                self._problem_begin_idx = _ptr
                break
            _ptr += len(_word)

    def _set_end_idx(self, problem, answer_list):
        # right to left retrieval
        _ptr = len(problem)
        for _idx, _answer in reversed(list(enumerate(answer_list))):
            _word, _pos = _answer
            if _ptr - len(_word) < 0:
                self._answer_end_idx = _idx
                self._problem_end_idx = _ptr
                break

            _problem_sub_token = problem[_ptr - len(_word):_ptr]
            if _word != _problem_sub_token:
                self._answer_end_idx = _idx
                self._problem_end_idx = _ptr
                break

            _ptr = _ptr - len(_word)

    def _get_irregular_rules(self, problem, answer_list):
        _irr_rules = []
        if not self._is_clean_rule(problem, answer_list):
            return _irr_rules
        # 불규칙 시작지점과 끝지점이 cross 되는 경우
        # 알리가 -> 알 + ㄹ + 리가
        # 시작지점 ( left to right ) = 4 ( ㅣㄱㅏ )
        # 끝 지점 ( right to left ) = 2 ( ㅇㅏ )
        if self._problem_begin_idx > self._problem_end_idx:
            # 시작 지점 처리
            _irr_problem = problem[0:self._problem_end_idx].strip()
            if len(_irr_problem) == 0 or self._answer_end_idx == 0:
                _irr_rule = self._expand_irregular_rule(problem, answer_list, 0, self._problem_end_idx, 0,
                                                        self._answer_end_idx)
            else:
                _irr_rule = self._build_irregular_rule(_irr_problem, answer_list, 0, self._answer_end_idx)
            _irr_rules.append(_irr_rule)

            # 끝 지점 처리
            _irr_problem = problem[self._problem_begin_idx:].strip()
            if len(_irr_problem) == 0 or self._answer_begin_idx == len(answer_list) - 1:
                if self._answer_begin_idx - 1 >= len(answer_list):
                    print(f"{problem} -> {answer_list}")
                    print(answer_list[self._answer_begin_idx - 1])
                _irr_rule = self._expand_irregular_rule(problem, answer_list, self._problem_begin_idx, len(problem),
                                                        self._answer_begin_idx, len(answer_list) - 1)
            else:
                _irr_rule = self._build_irregular_rule(_irr_problem, answer_list, self._answer_begin_idx,
                                                       len(answer_list) - 1)
            _irr_rules.append(_irr_rule)
        # 축약인 경우
        # 너라고 -> 너 + 이 + 라고
        # 불규칙 패턴의 형태소룰 앞, 뒤로 하나씩 확장
        # "null" -> "이"의 패턴을 "너라고 -> 너 + 이 + 라고"로 확장
        elif self._problem_begin_idx == self._problem_end_idx:
            _irr_rule = self._expand_irregular_rule(problem, answer_list, self._problem_begin_idx,
                                                    self._problem_end_idx, self._answer_begin_idx, self._answer_end_idx)
            _irr_rules.append(_irr_rule)
        else:
            _irr_problem = problem[self._problem_begin_idx:self._problem_end_idx].strip()
            # 탈락인 경우
            # 어떠냐 -> 어떻 + 냐
            # 어떠 -> null 인 경우 단일 형태소 그대로 사용 ( 어떠 -> 어떻 )
            if self._answer_begin_idx == self._answer_end_idx:
                _extract_length = self._problem_end_idx - self._problem_begin_idx
                if _extract_length < 4:
                    _irr_rule = self._expand_irregular_rule(problem, answer_list, self._problem_begin_idx,
                                                            self._problem_end_idx, self._answer_begin_idx,
                                                            self._answer_end_idx)
                else:
                    _irr_rule = self._build_irregular_rule(_irr_problem, answer_list, self._answer_begin_idx,
                                                           self._answer_end_idx)
            else:
                _irr_rule = self._build_irregular_rule(_irr_problem, answer_list, self._answer_begin_idx,
                                                       self._answer_end_idx)
            _irr_rules.append(_irr_rule)

        return _irr_rules

    def _is_clean_rule(self, problem, answer_list):

        if self._answer_begin_idx == -1 or self._answer_end_idx == -1:
            return False

        if answer_list[self._answer_end_idx] == "VCP":
            return False

        i = self._answer_begin_idx
        while i < self._answer_end_idx+1:
            _pos = answer_list[i][1]
            if _pos == "NNP" or _pos == "NNG" or _pos.startswith("JK"):
                return False
            i += 1

        for _letter in problem:
            if not is_korean(_letter):
                return False

        return True

    @staticmethod
    def _expand_irregular_rule(problem, answer_list, problem_begin_idx, problem_end_idx, answer_begin_idx,
                               answer_end_idx):
        _center_convert_rule = ""
        for i in range(answer_begin_idx, answer_end_idx + 1):
            _word = answer_list[i][0]
            _pos = answer_list[i][1]
            _center_convert_rule += f"{_word}/{_pos} "

        _prev_convert_rule = ""
        _next_convert_rule = ""

        _tmp_problem_begin_idx = problem_begin_idx
        _tmp_problem_end_idx = problem_end_idx

        if answer_begin_idx != 0:
            _word = answer_list[answer_begin_idx - 1][0]
            _pos = answer_list[answer_begin_idx - 1][1]
            _prev_convert_rule += f"{_word}/{_pos} "
            _tmp_problem_begin_idx = problem_begin_idx - len(_word)

        if answer_end_idx + 1 != len(answer_list):
            _word = answer_list[answer_end_idx + 1][0]
            _pos = answer_list[answer_end_idx + 1][1]
            _next_convert_rule += f"{_word}/{_pos} "
            _tmp_problem_end_idx = problem_end_idx + len(_word)

        _convert_rule = _prev_convert_rule + _center_convert_rule + _next_convert_rule.strip()
        _irr_problem = problem[_tmp_problem_begin_idx:_tmp_problem_end_idx]
        return _irr_problem, _convert_rule.strip()

    @staticmethod
    def _build_irregular_rule(problem, answer_list, answer_begin_idx, answer_end_idx):
        _convert_rule = ""
        for i in range(answer_begin_idx, answer_end_idx + 1):
            _word = answer_list[i][0]
            _pos = answer_list[i][1]
            _convert_rule += f"{_word}/{_pos} "
        return problem, _convert_rule.strip()


import os
import re
import shutil

from ..common.dictionary import Dictionary
from ..common.grammar import Grammar
from ..constant import SYMBOL, FILENAME
from ..parser.corpus_parser import CorpusParser
from ..parser.irregular_parser import IrregularParser
from ..parser.korean_unit_parser import KoreanUnitParser, is_jamo


class CorpusBuilder:
    def __init__(self):
        self.unit_parser = KoreanUnitParser()
        self.corpus_parser = CorpusParser()
        self.irregular_parser = IrregularParser()
        self.word_dic = Dictionary()
        self.irr_dic = Dictionary()
        self.grammar = Grammar()

    def save(self, path):
        if os.path.exists(path):
            shutil.rmtree(path)
        os.makedirs(path)
        self.word_dic.save(path + "/" + FILENAME.WORD_DIC)
        self.irr_dic.save(path + "/" + FILENAME.IRREGULAR_DIC)
        self.grammar.save(path + "/" + FILENAME.GRAMMAR)

    def load(self, path):
        self.word_dic.load(path + "/" + FILENAME.WORD_DIC)
        self.irr_dic.load(path + "/" + FILENAME.IRREGULAR_DIC)
        self.grammar.load(path + "/" + FILENAME.GRAMMAR)

    def build_path(self, corpora_path, suffix=None):
        _filenames = [f.name for f in os.scandir(corpora_path) if f.is_file()]
        for _filename in _filenames:
            _abs_filename = corpora_path + "/" + _filename
            if suffix is None:
                print(_abs_filename)
                self.build(_abs_filename)
            else:
                if _filename.endswith(suffix):
                    print(_abs_filename)
                    self.build(_abs_filename)
        print("before pruning")
        print(f"dic size = {len(self.word_dic.get_dict())}")
        print(f"irr dic size = {len(self.irr_dic.get_dict())}")
        print(f"grammar size = {len(self.grammar.get_grammar())}")
        self._pruning()
        print("after pruning")
        print(f"dic size = {len(self.word_dic.get_dict())}")
        print(f"irr dic size = {len(self.irr_dic.get_dict())}")
        print(f"grammar size = {len(self.grammar.get_grammar())}")

    def build(self, _filename):
        with open(_filename, 'r') as f:
            for line in f.readlines():
                line = re.sub("[ ]+", " ", line).strip()
                if len(line) == 0:
                    continue
                pa_pair = self.corpus_parser.parse(line)
                if pa_pair is None:
                    continue
                self._append_word_dic(pa_pair.get_answer_list())
                self._append_irr_dic(pa_pair)
                self._append_grammar(pa_pair.get_answer_list())

    def _append_word_dic(self, answer_list):
        for answer in answer_list:
            # todo : KOMORAN 에는 뭔가 예외처리가 있는 것으로 보이는데 그 부분은 구현하지 않았음
            _word = answer[0]
            _pos = answer[1]
            self.word_dic.append(_word, _pos)

    def _append_irr_dic(self, problem_answer):
        if self._is_irregular(problem_answer.get_problem(), problem_answer.get_answer_list()):
            _irr_rules = self.irregular_parser.parse(KoreanUnitParser.parse(problem_answer.get_problem()),
                                                     [(KoreanUnitParser.parse(_word), _pos) for _word, _pos in
                                                      problem_answer.get_answer_list()])
            for _irr_rule in _irr_rules:
                _irr = _irr_rule[0]
                _rule = _irr_rule[1]
                if len(_rule.strip()) == 0:
                    continue
                # 불규칙 대상에 자소가 하나의 음절로 포함된 경우 skip
                # todo : KOMORAN에서는 불규칙 예외에 대한 처리를 하지만 그 부분은 구현하지 않았음
                _has_jaso_problem = False
                reunion_irr_syllable = KoreanUnitParser.combine(_irr)
                reunion_rule_syllable = KoreanUnitParser.combine(_rule)
                for _letter in reunion_irr_syllable:
                    if is_jamo(_letter):
                        _has_jaso_problem = True
                        break
                if _has_jaso_problem:
                    continue
                self.irr_dic.append(reunion_irr_syllable, reunion_rule_syllable)

    def _is_irregular(self, problem, answer_list):
        _answer_str = ""
        for answer in answer_list:
            _word = answer[0]
            _answer_str += _word
        problem_unit = self.unit_parser.parse(problem)
        answer_unit = self.unit_parser.parse(_answer_str)
        # todo : KOMORAN 에서는 한국어만 추출해서 비교하지만 그 부분은 생략하였음
        return problem_unit != answer_unit

    def _append_grammar(self, answer_list):
        _prev_pos = SYMBOL.BOE
        for answer in answer_list:
            # todo : KOMORAN에서는 correctGrammar라는 별도의 문법을 입력 받은 뒤에 처리하는 로직이 있으나 구현하지 않음
            _current_pos = answer[1]
            self.grammar.append(_prev_pos, _current_pos)
            _prev_pos = answer[1]
        _end_pos = SYMBOL.EOE
        self.grammar.append(_prev_pos, _end_pos)

    def _pruning(self):
        # dic pruning
        pruned_dic, pruned_words = self._pruning_dic(self.word_dic.get_dict(), freq_threshold=2)
        self.word_dic = Dictionary(pruned_dic)
        with open('excluded_nouns.txt', 'w') as f:
            for word in pruned_words:
                f.write(word + "\n")
        # grammar pruning
        self.grammar = Grammar(self._pruning_dic(self.grammar.get_grammar())[0])
        # irregular dic pruning
        self.irr_dic = Dictionary(self._pruning_dic(self.irr_dic.get_dict())[0])

    @staticmethod
    def _pruning_dic(dic, freq_threshold=5):
        _pruned_dic = {}
        _pruned_words = []
        for _word, _pos_freq_dic in dic.items():
            _pruned_pos_freq_dic = {}
            for _pos, _freq in _pos_freq_dic.items():
                if _freq <= freq_threshold:
                    if _pos == 'NNG' or _pos == 'NNP':
                        _pruned_words.append(_word)
                    continue
                _pruned_pos_freq_dic[_pos] = _freq
            if len(_pruned_pos_freq_dic) != 0:
                _pruned_dic[_word] = _pruned_pos_freq_dic
        return _pruned_dic, _pruned_words



import numpy as np
import pickle
import gzip


class Transition:
    def __init__(self, size):
        self.score_matrix = np.full((size, size), -np.inf)

    def put(self, prev_pos_id, cur_pos_id, score):
        self.score_matrix[prev_pos_id][cur_pos_id] = score

    def get(self, prev_pos_id, cur_pos_id):
        score = self.score_matrix[prev_pos_id][cur_pos_id]
        # has_transition_score = np.where(score == -np.inf, False, True)
        # print(f"{prev_pos_id}->{cur_pos_id}:{score}")
        # print(has_transition_score)
        # if has_transition_score:
        #     return score
        # else:
        #     return None
        if score == -np.inf:
            return None
        return score

    def save(self, filename):
        with gzip.open(filename, 'wb') as f:
            pickle.dump(self.score_matrix, f)

    def load(self, filename):
        with gzip.open(filename, 'rb') as f:
            self.score_matrix = pickle.load(f)



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="komoranPi",
    version="0.0.1",
    author="Junsoo Shin",
    author_email="ceo@shineware.co.kr",
    description="KOMORAN의 native python 버전입니다.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/shineware/komoranpy",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
)


class Grammar:
    def __init__(self, init_grammar=None):
        if init_grammar is None:
            init_grammar = {}
        self._grammar = init_grammar

    def load(self, filename):
        self._grammar = {}
        with open(filename, 'r') as f:
            for line in f.readlines():
                line = line.strip()
                if len(line) == 0:
                    continue
                _tokens = line.split('\t')
                _prev_pos = _tokens[0]
                _cur_pos_freq_dic = {}
                _cur_pos_tokens = _tokens[1].split(',')
                for _token_idx in range(0, len(_cur_pos_tokens)):
                    _token = _cur_pos_tokens[_token_idx]
                    _separator_idx = _token.rindex(':')
                    _cur_pos = _token[0:_separator_idx]
                    _freq = int(_token[_separator_idx + 1:])
                    _cur_pos_freq_dic[_cur_pos] = _freq
                self._grammar[_prev_pos] = _cur_pos_freq_dic

    def save(self, filename):
        with open(filename, 'w') as f:
            for _prev_pos, _cur_pos_freq_dic in self._grammar.items():
                f.write(
                    _prev_pos + "\t" + ",".join(
                        [f"{_cur_pos}:{_freq}" for _cur_pos, _freq in _cur_pos_freq_dic.items()]) + "\n")

    def append(self, prev_pos, cur_pos, freq=1):
        _cur_pos_freq_dic = self._grammar.get(prev_pos)
        if _cur_pos_freq_dic is None:
            _cur_pos_freq_dic = {}

        _freq = _cur_pos_freq_dic.get(cur_pos)
        if _freq is None:
            _freq = 0
        _freq += freq
        _cur_pos_freq_dic[cur_pos] = _freq
        self._grammar[prev_pos] = _cur_pos_freq_dic

    def print(self):
        print(self._grammar)

    def __repr__(self):
        return f'Grammar: {self._grammar}'

    def has(self, prev_pos, cur_pos):
        _cur_pos_freq_dict = self._grammar.get(prev_pos)
        if _cur_pos_freq_dict is None:
            return False
        else:
            _freq = _cur_pos_freq_dict.get(cur_pos)
            return _freq is not None

    def get_grammar(self):
        return self._grammar



import gzip
import pickle

from .ahocorasick_node import AhoCorasickNode


class FindContext:
    def __init__(self, init_node):
        self.current_node = init_node

    def set_current_node(self, node):
        self.current_node = node

    def get_current_node(self):
        return self.current_node

    def get_current_fail_node(self):
        return self.current_node.get_fail_node()

    def get_current_child_nodes(self):
        return self.current_node.get_child_nodes()

    def is_root(self):
        return self.current_node.get_parent_node() is None


class AhoCorasickDictionary:
    def __init__(self):
        self.root_node = AhoCorasickNode()
        self.root_node.set_depth(0)

    def put(self, keys, value):
        current_node = self.root_node
        for depth, key in enumerate(keys):
            child_nodes = current_node.get_child_nodes()
            child_node = child_nodes.get(key)
            if child_node is None:
                node = AhoCorasickNode()
                node.set_parent_node(current_node)
                node.set_depth(depth + 1)
                node.set_key(key)
                child_nodes[key] = node
                current_node.set_child_nodes(child_nodes)
            current_node = current_node.get_child_nodes()[key]
        current_node.set_value(value)

    def get(self, key, context):
        result_dic = {}
        while True:
            child_nodes = context.get_current_child_nodes()
            if len(child_nodes) == 0:
                fail_node = context.get_current_fail_node()
                if fail_node is None:
                    return None
                else:
                    context.set_current_node(fail_node)
                continue
            child_node = child_nodes.get(key)
            if child_node is None:
                if context.get_current_node() != self.root_node:
                    context.set_current_node(context.get_current_fail_node())
                    continue
            else:
                if child_node.get_value() is not None:
                    result_dic[self.get_key_from_root(child_node)] = child_node.get_value()
                while child_node.get_fail_node() is not None:
                    if child_node.get_fail_node().get_value() is not None:
                        result_dic[
                            self.get_key_from_root(child_node.get_fail_node())] = child_node.get_fail_node().get_value()
                    child_node = child_node.get_fail_node()
                context.set_current_node(child_nodes.get(key))
            break
        return result_dic

    def get_value(self, keys):
        node = self.root_node
        for key in keys:
            child_nodes = node.get_child_nodes()
            node = child_nodes.get(key)
            if node is None:
                return None
        return node.get_value()

    def build_fail_link(self):
        current_node = self.root_node
        queue = [current_node]

        while len(queue) != 0:
            current_node = queue.pop(0)
            self._link_fail_node(current_node)
            for node in current_node.get_child_nodes():
                queue.append(current_node.get_child_nodes()[node])

    def _link_fail_node(self, current_node):
        if current_node == self.root_node:
            pass
        elif current_node.get_parent_node() == self.root_node:
            current_node.set_fail_node(self.root_node)
        else:
            fail_node = current_node.get_parent_node().get_fail_node()
            while fail_node != self.root_node:
                if len(fail_node.get_child_nodes()) == 0:
                    fail_node = fail_node.get_fail_node()
                    continue
                child_node = fail_node.get_child_nodes().get(current_node.get_key())
                if child_node is not None:
                    current_node.set_fail_node(fail_node.get_child_nodes().get(current_node.get_key()))
                    break
                fail_node = fail_node.get_fail_node()
            if current_node.get_fail_node() is None:
                child_node = self.root_node.get_child_nodes().get(current_node.get_key())
                if child_node is not None:
                    current_node.set_fail_node(child_node)
                else:
                    current_node.set_fail_node(self.root_node)

    def new_find_context(self):
        return FindContext(self.root_node)

    def get_key_from_root(self, child_node):
        current_node = child_node
        key = ""
        while current_node != self.root_node:
            key = current_node.get_key() + key
            current_node = current_node.get_parent_node()
        return key

    def save(self, filename):
        with gzip.open(filename, 'wb') as f:
            pickle.dump(self.root_node, f)

    def load(self, filename):
        with gzip.open(filename, 'rb') as f:
            self.root_node = pickle.load(f)



import re
from bisect import bisect_left
from enum import Enum

chosung_list = [0x3131, 0x3132, 0x3134, 0x3137, 0x3138,
                0x3139, 0x3141, 0x3142, 0x3143, 0x3145, 0x3146, 0x3147, 0x3148,
                0x3149, 0x314a, 0x314b, 0x314c, 0x314d, 0x314e]
jungsung_list = [0x314f, 0x3150, 0x3151, 0x3152, 0x3153,
                 0x3154, 0x3155, 0x3156, 0x3157, 0x3158, 0x3159, 0x315a, 0x315b,
                 0x315c, 0x315d, 0x315e, 0x315f, 0x3160, 0x3161, 0x3162, 0x3163]
jongsung_list = [0x0000, 0x3131, 0x3132, 0x3133, 0x3134,
                 0x3135, 0x3136, 0x3137, 0x3139, 0x313a, 0x313b, 0x313c, 0x313d,
                 0x313e, 0x313f, 0x3140, 0x3141, 0x3142, 0x3144, 0x3145, 0x3146,
                 0x3147, 0x3148, 0x314a, 0x314b, 0x314c, 0x314d, 0x314e]


class UnitType(Enum):
    CHOSUNG = 1
    JUNGSUNG = 2
    JONGSUNG = 3
    OTHER = 4


def binary_search(collection, element):
    _idx = bisect_left(collection, element)
    if _idx == len(collection) or collection[_idx] != element:
        return -1
    return _idx


class KoreanUnitParser:

    @staticmethod
    def parse_with_type(text):
        _result = []
        for ch in text:
            hex_value = ord(ch)
            # 음절 영역
            if 0xAC00 <= hex_value <= 0xD7AF:
                _cho_idx = _jung_idx = _jong_idx = _tmp_idx = 0
                _tmp_idx = int(hex_value - 0xAC00)
                _cho_idx = int(_tmp_idx / (21 * 28))
                _tmp_idx = int(_tmp_idx % (21 * 28))
                _jung_idx = int(_tmp_idx / 28)
                _jong_idx = int(_tmp_idx % 28)
                _result.append((chr(chosung_list[_cho_idx]), chosung_list[_cho_idx], UnitType.CHOSUNG))
                _result.append((chr(jungsung_list[_jung_idx]), jungsung_list[_jung_idx], UnitType.JUNGSUNG))
                if _jong_idx != 0:
                    _result.append((chr(jongsung_list[_jong_idx]), jongsung_list[_jong_idx], UnitType.JONGSUNG))

            else:
                _result.append((ch, None, UnitType.OTHER))
        return _result

    @staticmethod
    def parse(text):
        _result = []
        for ch in text:
            hex_value = ord(ch)
            # 음절 영역
            if 0xAC00 <= hex_value <= 0xD7AF:
                _cho_idx = _jung_idx = _jong_idx = _tmp_idx = 0
                _tmp_idx = int(hex_value - 0xAC00)
                _cho_idx = int(_tmp_idx / (21 * 28))
                _tmp_idx = int(_tmp_idx % (21 * 28))
                _jung_idx = int(_tmp_idx / 28)
                _jong_idx = int(_tmp_idx % 28)
                _result.append(chr(chosung_list[_cho_idx]))
                _result.append(chr(jungsung_list[_jung_idx]))
                if _jong_idx != 0:
                    _result.append(chr(jongsung_list[_jong_idx]))
            else:
                _result.append(ch)
        return "".join(_result)

    @staticmethod
    def combine_with_type(jaso_with_types):
        _chosung = _jungsung = _jongsung = 0
        _has_buffer = False

        _result = ""

        for _jaso, _hex, _type in jaso_with_types:
            if _type == UnitType.CHOSUNG:
                if _has_buffer:
                    _result += chr(0xac00 + _chosung * 588 + _jungsung * 28 + _jongsung)
                    _jungsung = _jongsung = 0
                _chosung = binary_search(chosung_list, _hex)
                _has_buffer = True
            elif _type == UnitType.JUNGSUNG:
                _jungsung = binary_search(jungsung_list, _hex)
                _has_buffer = True
            elif _type == UnitType.JONGSUNG:
                _jongsung = binary_search(jongsung_list, _hex)
                _has_buffer = True
            else:
                if _has_buffer:
                    _result += chr(0xac00 + _chosung * 588 + _jungsung * 28 + _jongsung)
                    _chosung = _jungsung = _jongsung = 0
                _result += _jaso
                _has_buffer = False

        if _has_buffer:
            _result += chr(0xac00 + _chosung * 588 + _jungsung * 28 + _jongsung)
        return _result

    @staticmethod
    def combine(jaso_text):
        _idx = 1
        _prev_idx = 0
        _result = ""
        while _idx < len(jaso_text):
            _hex = ord(jaso_text[_idx])
            _jungsung = binary_search(jungsung_list, _hex)
            if _jungsung >= 0:
                _hex = ord(jaso_text[_idx - 1])
                _chosung = binary_search(chosung_list, _hex)
                if _chosung < 0:
                    _idx += 1
                    continue
                _result += jaso_text[_prev_idx:_idx - 1]

                _jongsung = 0
                if _idx + 1 < len(jaso_text):
                    _hex = ord(jaso_text[_idx + 1])
                    _jongsung = binary_search(jongsung_list, _hex)
                if _idx + 2 < len(jaso_text) and binary_search(jungsung_list, ord(jaso_text[_idx + 2])) >= 0:
                    _jongsung = 0
                # 종성이 없는 경우
                if _jongsung < 0:
                    _jongsung = 0
                _result += chr(0xac00 + _chosung * 588 + _jungsung * 28 + _jongsung)
                if _jongsung > 0:
                    _idx += 1
                _prev_idx = _idx + 1
            _idx += 1
        if _prev_idx < len(jaso_text):
            _result += jaso_text[_prev_idx:]

        return _result


def is_korean(letter):
    return re.search('[ㄱ-ㅎ|ㅏ-ㅣ|가-힣]+', letter) is not None


def is_jamo(letter):
    _hex = ord(letter)
    return 0x3131 <= _hex <= 0x318E
